"""File containing constants used by Configurations."""

CONFIG_PATH_KEY = "MF_CONFIG_DIR"
CONFIG_DWH_DIALECT = "dwh_dialect"
CONFIG_DWH_HOST = "dwh_host"
CONFIG_DWH_PORT = "dwh_port"
CONFIG_DWH_USER = "dwh_user"
CONFIG_DWH_PASSWORD = "dwh_password"
CONFIG_DWH_DB = "dwh_database"
CONFIG_DWH_SCHEMA = "dwh_schema"
CONFIG_DWH_CREDS_PATH = "dwh_path_to_creds"
CONFIG_DWH_WAREHOUSE = "dwh_warehouse"
CONFIG_DWH_PROJECT_ID = "dwh_project_id"
CONFIG_EMAIL = "email"
CONFIG_MODEL_PATH = "model_path"
CONFIG_DWH_HTTP_PATH = "dwh_http_path"
CONFIG_DWH_ACCESS_TOKEN = "dwh_access_token"
CONFIG_DBT_REPO = "dbt_repo"
CONFIG_DBT_PROFILE = "dbt_profile"
CONFIG_DBT_TARGET = "dbt_target"
